@@ -1,15 +1,37 @@
 /* Compatibility routines.
  *
- * Services is copyright (c) 1996-1999 Andrew Church.
- *     E-mail: <achurch@dragonfire.net>
- * Services is copyright (c) 1999-2000 Andrew Kempe.
- *     E-mail: <theshadow@shadowfire.org>
+ * IRC Services is copyright (c) 1996-2007 Andrew Church.
+ *     E-mail: <achurch@achurch.org>
+ * Parts written by Andrew Kempe and others.
  * This program is free but copyrighted software; see the file COPYING for
  * details.
  */
 
 #include "services.h"
 
+#if !HAVE_HSTRERROR
+# include <netdb.h>
+#endif
+
+/*************************************************************************/
+
+#if !HAVE_HSTRERROR
+
+/* hstrerror: return an error message for the given h_errno code. */
+
+const char *hstrerror(int h_errnum)
+{
+    switch (h_errnum) {
+	case HOST_NOT_FOUND: return "Host not found";
+	case TRY_AGAIN:      return "Host not found, try again";
+	case NO_RECOVERY:    return "Nameserver error";
+	case NO_DATA:        return "No data of requested type";
+	default:             return "Unknown error";
+    }
+}
+
+#endif /* !HAVE_HSTRERROR */
+
 /*************************************************************************/
 
 #if !HAVE_SNPRINTF
@@ -36,15 +58,49 @@ int vsnprintf(char *buf, size_t len, const char *fmt, va_list args)
 int snprintf(char *buf, size_t len, const char *fmt, ...)
 {
     va_list args;
+    int ret;
 
     va_start(args, fmt);
-    return vsnprintf(buf, len, fmt, args);
+    ret = vsnprintf(buf, len, fmt, args);
+    va_end(args);
+    return ret;
 }
 
 #endif /* !HAVE_SNPRINTF */
 
 /*************************************************************************/
 
+#if !HAVE_STRTOK
+
+/* glibc 2.2 (RedHat 7.0) has a broken strtok--it dies if called with a
+ * NULL parameter after returning NULL once.  glibc and possibly other
+ * libraries also return non-NULL for strtok(NULL, "") even after
+ * returning NULL for strtok(NULL, " ").
+ */
+
+char *strtok(char *str, const char *delim)
+{
+    static char *current = NULL;
+    char *ret;
+
+    if (str)
+	current = str;
+    if (!current)
+	return NULL;
+    current += strspn(current, delim);
+    ret = *current ? current : NULL;
+    current += strcspn(current, delim);
+    if (!*current)
+	current = NULL;
+    else
+	*current++ = 0;
+    return ret;
+}
+
+#endif /* !HAVE_STRTOK */
+
+/*************************************************************************/
+
 #if !HAVE_STRICMP && !HAVE_STRCASECMP
 
 /* stricmp, strnicmp:  Case-insensitive versions of strcmp() and
@@ -56,8 +112,10 @@ int stricmp(const char *s1, const char *s2)
     register int c;
 
     while ((c = tolower(*s1)) == tolower(*s2)) {
-	if (c == 0) return 0;
-	s1++; s2++;
+	if (c == 0)
+	    return 0;
+	s1++;
+	s2++;
     }
     if (c < tolower(*s2))
 	return -1;
@@ -68,10 +126,13 @@ int strnicmp(const char *s1, const char *s2, size_t len)
 {
     register int c;
 
-    if (!len) return 0;
+    if (!len)
+	return 0;
     while ((c = tolower(*s1)) == tolower(*s2) && len > 0) {
-	if (c == 0 || --len == 0) return 0;
-	s1++; s2++;
+	if (c == 0 || --len == 0)
+	    return 0;
+	s1++;
+	s2++;
     }
     if (c < tolower(*s2))
 	return -1;
@@ -94,6 +155,7 @@ char *strdup(const char *s)
 /*************************************************************************/
 
 #if !HAVE_STRSPN
+
 size_t strspn(const char *s, const char *accept)
 {
     size_t i = 0;
@@ -102,6 +164,16 @@ size_t strspn(const char *s, const char *accept)
 	++i, ++s;
     return i;
 }
+
+size_t strcspn(const char *s, const char *reject)
+{
+    size_t i = 0;
+
+    while (*s && !strchr(reject, *s))
+	++i, ++s;
+    return i;
+}
+
 #endif
 
 /*************************************************************************/
