@@ -1,17 +1,20 @@
 /* Routines for time-delayed actions.
  *
- * Services is copyright (c) 1996-1999 Andrew Church.
- *     E-mail: <achurch@dragonfire.net>
- * Services is copyright (c) 1999-2000 Andrew Kempe.
- *     E-mail: <theshadow@shadowfire.org>
+ * IRC Services is copyright (c) 1996-2007 Andrew Church.
+ *     E-mail: <achurch@achurch.org>
+ * Parts written by Andrew Kempe and others.
  * This program is free but copyrighted software; see the file COPYING for
  * details.
  */
 
 #include "services.h"
+#define IN_TIMEOUT_C
 #include "timeout.h"
 
+/*************************************************************************/
+
 static Timeout *timeouts = NULL;
+static int checking_timeouts = 0;
 
 /*************************************************************************/
 
@@ -21,16 +24,13 @@ static Timeout *timeouts = NULL;
 
 void send_timeout_list(User *u)
 {
-    Timeout *to, *last;
+    Timeout *to;
+    uint32 now = time_msec();
 
-    notice(s_OperServ, u->nick, "Now: %ld", time(NULL));
-    for (to = timeouts, last = NULL; to; last = to, to = to->next) {
-	notice(s_OperServ, u->nick, "%p: %ld: %p (%p)",
-			to, to->timeout, to->code, to->data);
-	if (to->prev != last)
-	    notice(s_OperServ, u->nick,
-			"    to->prev incorrect!  expected=%p seen=%p",
-			last, to->prev);
+    notice(ServerName, u->nick, "Now: %u.%03u", now/1000, now%1000);
+    LIST_FOREACH (to, timeouts) {
+	notice(ServerName, u->nick, "%p: %u.%03u: %p (%p)",
+	       to, to->timeout/1000, to->timeout%1000, to->code, to->data);
     }
 }
 
@@ -43,38 +43,37 @@ void send_timeout_list(User *u)
 void check_timeouts(void)
 {
     Timeout *to, *to2;
-    time_t t = time(NULL);
+    uint32 now = time_msec();
 
-    if (debug >= 2)
-	log("debug: Checking timeouts at %ld", t);
-
-    to = timeouts;
-    while (to) {
-	if (t < to->timeout) {
-	    to = to->next;
-	    continue;
-	}
-	if (debug >= 4) {
-	    log("debug: Running timeout %p (code=%p repeat=%d)",
-			to, to->code, to->repeat);
-	}
-	to->code(to);
-	if (to->repeat) {
-	    to = to->next;
-	    continue;
-	}
-	to2 = to->next;
-	if (to->next)
-	    to->next->prev = to->prev;
-	if (to->prev)
-	    to->prev->next = to->next;
-	else
-	    timeouts = to->next;
-	free(to);
-	to = to2;
+    if (checking_timeouts)
+	fatal("check_timeouts() called recursively!");
+    checking_timeouts = 1;
+    if (debug >= 2) {
+	log("debug: Checking timeouts at time_msec = %u.%03u",
+	    now/1000, now%1000);
     }
+
+    LIST_FOREACH_SAFE (to, timeouts, to2) {
+	if (to->timeout) {
+	    if ((int32)(to->timeout - now) > 0)
+		continue;
+	    if (debug >= 3) {
+		log("debug: Running timeout %p (code=%p repeat=%d)",
+		    to, to->code, to->repeat);
+	    }
+	    to->code(to);
+	    if (to->repeat) {
+		to->timeout = now + to->repeat;
+		continue;
+	    }
+	}
+	LIST_REMOVE(to, timeouts);
+	free(to);
+    }
+
     if (debug >= 2)
 	log("debug: Finished timeout list");
+    checking_timeouts = 0;
 }
 
 /*************************************************************************/
@@ -86,17 +85,23 @@ void check_timeouts(void)
  */
 
 Timeout *add_timeout(int delay, void (*code)(Timeout *), int repeat)
+{
+    if (delay > 4294967)  /* 2^32 / 1000 */
+	delay = 4294967;
+    return add_timeout_ms((uint32)delay*1000, code, repeat);
+}
+
+/*************************************************************************/
+
+Timeout *add_timeout_ms(uint32 delay, void (*code)(Timeout *), int repeat)
 {
     Timeout *t = smalloc(sizeof(Timeout));
     t->settime = time(NULL);
-    t->timeout = t->settime + delay;
+    t->timeout = time_msec() + delay;
     t->code = code;
-    t->repeat = repeat;
-    t->next = timeouts;
-    t->prev = NULL;
-    if (timeouts)
-	timeouts->prev = t;
-    timeouts = t;
+    t->data = NULL;
+    t->repeat = repeat ? delay : 0;
+    LIST_INSERT(t, timeouts);
     return t;
 }
 
@@ -108,18 +113,19 @@ void del_timeout(Timeout *t)
 {
     Timeout *ptr;
 
-    for (ptr = timeouts; ptr; ptr = ptr->next) {
+    LIST_FOREACH (ptr, timeouts) {
 	if (ptr == t)
 	    break;
     }
-    if (!ptr)
+    if (!ptr) {
+	log("timeout: BUG: attempted to remove timeout %p (not on list)", t);
 	return;
-    if (t->prev)
-	t->prev->next = t->next;
-    else
-	timeouts = t->next;
-    if (t->next)
-	t->next->prev = t->prev;
+    }
+    if (checking_timeouts) {
+	t->timeout = 0;  /* delete it when we hit it in the list */
+	return;
+    }
+    LIST_REMOVE(t, timeouts);
     free(t);
 }
 
